'''
本库用于科学计算和快速建模

maysics主要包括十三个模块：

1、algorithm 封装了几种模拟方法，用于简易模拟；
2、calculus 封装了部分常见的算符算子和积分方法，辅助数学运算；
3、constant 储存了部分常数；
4、equation 封装了部分方程求解运算；
5、explainer 用于评估和解释模型；
6、graph 用于图论分析；
7、imageprocess 用于数字图像处理;
8、models 封装了几种常用的模型以便快速构建数理模型；
9、preprocess 用于数据预处理；
10、stats 用于统计分析；
11、tprocess 用于处理时间数据；
12、transformation 储存了常用的坐标转换及其他数学变换；
13、utils 是额外工具箱。


This package is used for scientific calculating and fast modeling.

maysics includes thirteen modules:

1. "algorithm" packages several simulation methods for simple simulation;
2. "calculus" packages some common operators and integration method to assist in mathematical operations;
3. "constant" contents some usual constants;
4. "equation" packages some equation solving operation;
5. "explainer" is used for estimating and explaining model;
6. "graph" is used for graph theory analysis;
7. "imageprocess" is used for digital image process;
8. "models" packages several commonly used models for fast modeling;
9. "preprocess" is used for data preproccessing;
10. "stats" is uesd for statistical analysis;
11. "tprocess" is used for processing time data;
12. "transformation" stores common coordinate transformations and other mathematical transformations;
13. "utils" is extra Utils.
'''
import numpy as np
import pickle, csv
from PIL import Image
from maysics import algorithm, calculus, constant, equation, explainer, graph,\
    models, preprocess, stats, tprocess, transformation, utils
from maysics.models import linear_r
from maysics.preprocess import preview, preview_file, shuffle
from maysics.utils import circle, discrete, grid_net


def covs1d(a, b, n):
    '''
    一维序列卷积和
    
    参数
    ----
    a：一维数组
    b：一维数组
    n：整型，平移步数
    
    返回
    ----
    数类型，a[n] * b[n]
    
    Convolution Sum of 1-D List
    
    Parameters
    ----------
    a: 1-D array
    b: 1-D array
    n: int, translation steps
    
    Return
    ------
    num, a[n] * b[n]
    '''
    a = np.array(a)
    b = list(b)
    b.reverse()
    b = np.array(b)
    num_a = len(a)
    num_b = len(b)
    if n <= 0 or n >= num_a + num_b:
        result = 0
    
    else:
        a = np.hstack((np.zeros(num_b), a, np.zeros(num_b)))
        b = np.hstack((b, np.zeros(num_a + num_b)))
        b[n : n+num_b] = b[: num_b]
        b[: n] = 0
        result = sum(a * b)
    
    return result


def covs2d(a, b, n, m):
    '''
    二维序列卷积和
    
    参数
    ----
    a：二维数组
    b：二维数组
    n：整型，沿axis=0方向的平移步数
    m：整型，沿axis=1方向的平移步数
    
    返回
    ----
    数类型，a[n, m] * b[n, m]
    
    Convolution Sum of 2-D List
    
    Parameters
    ----------
    a: 2-D array
    b: 2-D array
    n: int, translation steps along axis=0
    m: int, translation steps along axis=1
    
    Return
    ------
    num, a[n, m] * b[n, m]
    '''
    a = np.array(a)
    b = np.array(b)
    b = np.fliplr(b)
    b = np.flipud(b)
    num_a_x = a.shape[1]
    num_a_y = a.shape[0]
    num_b_x = b.shape[1]
    num_b_y = b.shape[0]
    
    if n <= 0 and m <= 0 or n >= num_a_x + num_b_x and m >= num_a_y + num_b_y:
        result = 0
    
    else:
        a = np.hstack((np.zeros((num_a_y, num_b_x)), a, np.zeros((num_a_y, num_b_x))))
        a = np.vstack((np.zeros((num_b_y, num_a_x + 2 * num_b_x)), a, np.zeros((num_b_y, num_a_x + 2 * num_b_x))))
        b = np.hstack((b, np.zeros((num_b_y, num_a_x + num_b_x))))
        b = np.vstack((b, np.zeros((num_a_y + num_b_y, num_a_x + 2 * num_b_x))))
        
        # 移动b矩阵
        b[n : n+num_b_y] = b[: num_b_y]
        b[: n] = 0
        b[:, m : m+num_b_x] = b[:, : num_b_x]
        b[:, : m] = 0
        
        result = (a * b).sum()
    
    return result


def save(filename, data, header=None):
    '''
    保存为.pkl、.npy或.csv文件
    
    参数
    ----
    filename：字符串类型，文件名
    data：需要保存的数据
    header：一维列表类型，可选，数据的列名称，仅在写入csv文件时有效
    
    
    Save as .pkl, .npy or .csv file
    
    Parameters
    ----------
    filename: str, file name
    data: data
    header: 1-D list, callable, the names of columns, effective only when writing csv files
    '''
    if filename[-4:] == '.pkl':
        with open(filename, 'wb') as file:
            pickle.dump(data, file)
    
    elif filename[-4:] == '.npy':
        np.save(filename, data)
    
    elif filename[-4:] == '.csv':
        data = np.array(data, dtype=np.object)
        if not header:
            header = []
            if len(data.shape) == 1:
                for i in range(data.shape[0]):
                    header.append(i)
                    with open(filename, 'w', newline='') as f:
                        writer = csv.writer(f)
                        writer.writerow(header)
                        writer.writerow(data)
            
            else:
                for i in range(data.shape[1]):
                    header.append(i)
                    with open(filename, 'w', newline='') as f:
                        writer = csv.writer(f)
                        writer.writerow(header)
                        writer.writerows(data)
    
    else:
        raise Exception("Suffix of filename must be '.pkl', '.npy' or '.csv'.")


def load(filename, header=True):
    '''
    载入.pkl、.npy或.csv文件
    
    参数
    ----
    filename：字符串类型，文件名
    header：布尔类型，可选，True表示csv文件第一行为列名，仅在读取csv文件时有效，默认为True
    
    
    Load .pkl, .npy or .csv file
    
    Parameter
    ---------
    filename: str, file name
    header: bool, callable, True means the first row of the csv file if the names of columns, effective only when reading csv files, default=True
    '''
    if filename[-4:] == '.pkl':
        with open(filename, 'rb') as file:
            data = pickle.load(file)
        
        return data
    
    elif filename[-4:] == '.npy':
        return np.load(filename, allow_pickle=True)
    
    elif filename[-4:] == '.csv':
        with open(filename, 'r') as f:
            reader = list(csv.reader(f))
            if header:
                reader = reader[1:]
            return np.array(reader)
    
    else:
        raise Exception("Suffix of filename must be '.pkl', '.npy' or '.csv'.")


def all_same(x):
    '''
    判断数组元素是否全相同
    
    参数
    ----
    x：数组
    
    返回
    ----
    布尔类型，True或者False
    
    
    Determine whether the array elements are all the same
    
    Parameter
    ---------
    x: array
    
    Return
    ------
    bool, True or False
    '''
    x = np.array(x)
    if len(x.shape) == 1:
        x = len(set(x))
        if x == 1:
            return True
        else:
            return False
    
    else:
        for i in x:
            if i.all() != x[0].all():
                return False
        return True


'''
本模块用于评估和选择模型

This module is ued for estimating and selecting models
'''
import numpy as np
from matplotlib import pyplot as plt
from scipy.special import comb
plt.rcParams['axes.unicode_minus'] =False


def _image_detail(data, labels, index, top, estimate, n1, n2):
    if not index:
        index = np.arange(0, data.shape[0], dtype=int)
    new_data = data[index]
    
    # 从大到小排序
    sort_list = np.argsort(new_data)
    new_data = np.sort(new_data)
    
    if not labels:
        labels = sort_list.astype(str)
    else:
        labels = np.array(labels, dtype=str)
        labels = labels[sort_list]
    
    if not top:
        top = data.shape[0]
    new_data = new_data[-top:]
    labels = labels[-top:]
    
    width = new_data.shape[0] * 0.5 * n1
    fig = plt.figure(figsize=(8, width))
    ax = fig.add_subplot(n1, 1, n2)
    
    color = []
    for i in range(new_data.shape[0]):
        if new_data[i] >= 0:
            color.append('#FFE100')
        else:
            color.append('#1E90FF')
    
    rows = ax.barh(range(new_data.shape[0]), new_data, color=color)
    for rect in rows:
        w = rect.get_width()
        if w >= 0:
            ha = 'left'
        else:
            ha = 'right'
        if not estimate:
            new_w = w
        else:
            new_w = np.around(w, estimate)
        ax.text(w, rect.get_y()+rect.get_height()/2, new_w, ha=ha, va='center')
    
    ax.set_yticks(range(new_data.shape[0]))
    ax.set_yticklabels(labels)
    plt.xticks(())


def _image_process(data, labels, index, top, estimate):
    if len(data.shape) == 1:
        _image_detail(data, labels, index, top, estimate, 1, 1)
    elif len(data.shape) == 2:
        num = data.shape[0]
        for j in range(num):
            _image_detail(data[j], labels, index, top, estimate, num, j+1)


def sense(func, x0, acc=0.1):
    '''
    灵敏度分析
    r = (x0, x1, x2, ..., xn)
    y = f(r)
    第i个特征在r=r0时的灵敏度：
    s(xi, r0)= (dy/dxi) * (xi/y)   (r=r0)
    
    参数
    ----
    func：函数类型，模型的预测函数，若函数需要输入列表，则列表须为ndarray
    x0：数，1-D或2-D ndarray，与func的输入格式相同，特征的初始值
    acc：浮点数类型，可选，求导的精度，默认为0.1
    
    返回
    ----
    各个维度的灵敏度，格式与func的输出格式相同
    
    
    Sensitivity Analysis
    r = (x0, x1, x2, ..., xn)
    y = f(r)
    the sensitivity of the ith feature at r=r0:
    s(xi, r0)= (dy/dxi) * (xi/y)   (r=r0)
    
    Parameters
    ----------
    func: function, predicting function of models, if the function requires a list as input, the list must be ndarray
    x0: num, 1-D or 2-D ndarray, the format is the same as the input of func, initial values of features
    acc: float, callable, accuracy of derivation, default=0.1
    
    Return
    ------
    sensitivity of each dimension, whose format is the same as the output of func
    '''
    x0 = np.array(x0, dtype=float)
    acc2 = 0.5 * acc
    func0 = func(x0)
    s_list = []
    
    if len(x0.shape) == 0:
        x0 += acc2
        func1 = func(x0)
        x0 -= acc
        func2 = func(x0)
        return (func1 - func2) / (acc * func0) * x0
    
    elif len(x0.shape) == 1:
        for i in range(x0.shape[0]):
            x0[i] += acc2
            func1 = func(x0)
            x0[i] -= acc
            func2 = func(x0)
            de = (func1 - func2) / (acc * func0) * x0[i]
            x0[i] += acc2
            s_list.append(de)
        s_list = np.array(s_list)
    
    elif len(x0.shape) == 2:
        for i in range(x0.shape[1]):
            x0[:, i] += acc2
            func1 = func(x0)
            x0[:, i] -= acc
            func2 = func(x0)
            de = (func1 - func2) / (acc * func0) * x0[:, i]
            x0[:, i] += acc2
            s_list.append(de)
        s_list = np.array(s_list).T
    return s_list


class Error():
    '''
    误差分析
    
    参数
    ----
    func：函数类型，模型的预测函数
    
    属性
    ----
    abs_error：1-D ndarray数组，绝对误差列表
    rel_error：1-D ndarray数组，相对误差列表
    abs_sort：元组列表，绝对误差从大到小的排序
    rel_sort：元组列表，相对误差从大到小的排序
    mae：平均绝对误差
    mape：平均绝对百分比误差
    mse：平均平方误差
    rmse：均方根误差
    sae：绝对误差和
    sse：残差平方和
    
    
    Error Analysis
    
    Parameter
    ---------
    func: function, predicting function of models
    
    Atrributes
    ----------
    abs_error: 1-D ndarray, absolute error list
    rel_error: 1-D ndarray, relative error list
    abs_sort: 1-D ndarray, list of absolute values of errors sorted from large to small
    rel_sort: 1-D ndarray, list of relative values of errors sorted from large to small
    mae: mean absolute error
    mape: mean absolute percentage error
    mse: mean squared error
    rmse: root mean square error
    sae: sum of absolute error
    sse: sum of squared error
    '''
    def __init__(self, func):
        self.__func = func
    
    
    def fit(self, data, target):
        '''
        计算误差
        
        参数
        ----
        data：ndarray，数据集的自变量
        target：ndarray，数据集的因变量
        
        
        Calculate the error
        
        Parameters
        ----------
        data: ndarray, independent variable of data set
        target: ndarray, dependent variable of data set
        '''
        data = np.array(data)
        target = np.array(target)
        self.__num_data = len(target)
        
        predict_target = self.__func(data)
        self.abs_error = abs(target - predict_target)
        self.rel_error = abs((target - predict_target) / target)
    
    @property
    def abs_sort(self):
        error_index = np.arange(self.__num_data)
        return sorted(list(zip(self.abs_error, error_index)), reverse=True)
    
    @property
    def rel_sort(self):
        error_index = np.arange(self.__num_data)
        return sorted(list(zip(self.rel_error, error_index)), reverse=True)
    
    @property
    def sse(self):
        return sum(self.abs_error**2)
    
    @property
    def sae(self):
        return sum(self.abs_error)
    
    @property
    def mse(self):
        return sum(self.abs_error**2) / self.__num_data
    
    @property
    def mae(self):
        return sum(self.abs_error) / self.__num_data
    
    @property
    def rmse(self):
        return (sum(self.abs_error**2) / self.__num_data)**0.5
    
    @property
    def mape(self):
        return sum(self.rel_error) / self.__num_data


class SHAP_and_Shapley():
    def __transform(self, b_list):
        # 二进制列表转十进制数
        d_value = 0
        for i in range(len(b_list)):
            d_value += b_list[i] * 2**i
        return d_value
    
    
    def _pre_fit(self, data, replace):
        dim = data.shape[1]
        shap_and_shapley_values = []
        
        # 求各个特征组合的预测值
        prediction = []
        for i in range(2**dim):
            new_data = data.copy()
            loc = np.array([i >>d & 1 for d in range(dim)][::1])
            loc = np.where(loc==0)[0]
            if len(loc) != 0:
                new_data[:, loc] = replace
            prediction_part = self.predict(new_data)
            prediction.append(prediction_part)
        prediction = np.array(prediction)
        
        # 求每个特征的SHAP值
        for i in range(dim):
            loc = np.array([2**i >>d & 1 for d in range(dim)][::1])
            loc = np.where(loc==1)[0][0]
            shap_value = 0
            
            # 每一层差值的加权求和
            for j in range(2**(dim-1)):
                b_list = [j >>d & 1 for d in range(dim)][::1]
                b_list.insert(loc, 0)
                b_value = self.__transform(b_list)
                # 1 / ((sum(b_list) + 1) * comb(dim, (sum(b_list) + 1)))是权重
                shap_value += (prediction[b_value + 2**i] - prediction[b_value]) / \
                              ((sum(b_list) + 1) * comb(dim, (sum(b_list) + 1)))
            shap_and_shapley_values.append(shap_value)
        return np.array(shap_and_shapley_values)
    
    
    def show(self, labels=None, index=None, top=None, estimate=None):
        '''
        作图并显示
        
        参数
        ----
        labels：一维列表，可选，特征名称，默认为None
        index：一维列表，可选，特征索引，默认为None，表示全选
        top：整型，可选，表示显示值最高的前top个特征，默认为None，表示全选
        estimate：整型，可选，表示图示示数保留的小数位数，默认为None
        
        
        Display the image
        
        Parameters
        ----------
        labels: 1-D list, callable, names of features, default=None
        index: 1-D list, callable, index of features, default=None, which means select all
        top: int, callable, display "top" features with the highest values, default=None, which means select all
        estimate: int, callable, indicating the number of decimal places reserved for the graphic display, default=None
        '''
        _image_process(self.values, labels, index, top, estimate)
        plt.show()
    
    
    def savefig(self, filename, labels=None, index=None, top=None, estimate=None):
        '''
        作图并保存
        
        参数
        ----
        filename：字符串类型，文件名
        labels：一维列表，可选，特征名称，默认为None
        index：一维列表，可选，特征索引，默认为None，表示全选
        top：整型，可选，表示显示值最高的前top个特征，默认为None，表示全选
        estimate：整型，可选，表示图示示数保留的小数位数，默认为None
        
        
        Save the image
        
        Parameters
        ----------
        filename: str, file name
        labels: 1-D list, callable, names of features, default=None
        index: 1-D list, callable, index of features, default=None, which means select all
        top: int, callable, display "top" features with the highest values, default=None, which means select all
        estimate: int, callable, indicating the number of decimal places reserved for the graphic display, default=None
        '''
        _image_process(self.values, labels, index, top, estimate)
        plt.savefig(filename)


class SHAP(SHAP_and_Shapley):
    '''
    计算局部点在模型中的SHAP值
    
    参数
    ----
    predict：函数类型，模型的预测函数
    
    属性
    ----
    values：一维ndarray，每个特征的SHAP值
    
    
    Calculate the SHAP of local points in the model
    
    Parameter
    ---------
    predict: function, the predict function of the model
    
    Attribute
    ---------
    values: 1-D ndarray, SHAP of each feature
    '''
    def __init__(self, predict):
        self.predict = predict
    
    
    def fit(self, data, replace=0):
        '''
        计算SHAP值
        
        参数
        ----
        data：一维数组，局部点
        replace：数或函数类型，可选，特征的替换值，函数须以np.array(data)为输入，默认为0
        
        
        Calculate the SHAP values
        
        Parameters
        ----------
        data: 1-D array, local point
        replace: num or function, callable, replacement values of features, for function, np.array(data) is the input, default=0
        '''
        data = np.array(data)
        # 确定替换值
        if type(replace).__name__ == 'function':
            replace = replace(data)
        data = np.array([data])
        
        self.values = SHAP_and_Shapley._pre_fit(self, data, replace).reshape(-1)


class Shapley(SHAP_and_Shapley):
    '''
    计算样本集在模型中的Shapley值
    
    参数
    ----
    predict：函数类型，模型的预测函数
    
    属性
    ----
    values：一维ndarray，每个特征的Shapley值
    
    
    Calculate the Shapley values of data set in the model
    
    Parameter
    ---------
    predict: function, the predict function of the model
    
    Attribute
    ---------
    values: 1-D ndarray, Shapley value of each feature
    '''
    def __init__(self, predict):
        self.predict = predict
    
    
    def fit(self, data, replace=0):
        '''
        计算Shapley
        
        参数
        ----
        data：二维数组，数据集
        replace：数或函数类型，可选，特征的替换值，函数须以np.array(data)为输入，默认为0
        
        
        Calculate the Shapley
        
        Parameters
        ----------
        data: 2-D array, data set
        replace: num or function, callable, replacement values of features, for function, np.array(data) is the input, default=0
        '''
        data = np.array(data)
        # 确定替换值
        if type(replace).__name__ == 'function':
            replace = replace(data)
        
        self.values = SHAP_and_Shapley._pre_fit(self, data, replace).T
        self.values = self.values.mean(axis=0)


class Lime():
    '''
    局部代理
    对局部点进行扰动并输入模型得到新的数据集
    以新数据集为基础，拟合出新的线性模型AX+b用于局部代替黑盒模型
    利用该线性模型对黑盒模型进行解释
    预测值 = Σweight + intercept
    
    参数
    ----
    predict_f：函数类型，原模型的预测函数，仅支持批量输入函数
    
    属性
    ----
    coef：ndarray，线性模型的系数
    intercept：数或一维ndarray，线性模型的截距
    weight：一维ndarray，各个特征对预测值的贡献
    
    
    Local Interpretable Model-agnostic Explanation
    perturb the local points and put them into the model to get a new data set
    fit a linear model Ax + b to replace the black box model base on the new data set
    explain the black box model with the linear model
    prediction = Σweight + intercept
    
    Parameter
    ---------
    predict_f: function, the predict function of the model, only batch input functions are supported
    
    Attributes
    ----------
    coef: ndarray, coeficcient of linear model
    intercept: num or 1-D ndarray, intercept of linear model
    weight: 1-D ndarray, contribution to the prediction of each feature
    '''
    def __init__(self, predict_f):
        self.predict_f = predict_f
    
    
    def fit(self, data, acc=0.1, num=100, random_state=None):
        '''
        进行Lime计算
        
        参数
        ----
        data：数组，局部点
        acc：浮点数类型，可选，邻域范围，默认为0.1
        num：整型，可选，在领域抽样点的数量，默认为100
        random_state：整型，可选，随机种子
        
        
        Calculate
        
        Parameters
        ----------
        data: ndarray, local point
        acc: float, callable, neighborhood range， default=0.1
        num: int, callable
        random_state: int, callable, random seed
        '''
        data = np.array(data)
        np.random.seed(random_state)
        X_set = np.random.rand(num, *data.shape) * 2 * acc - acc
        X_set += data
        y_set = self.predict_f(X_set)
        shape = X_set.shape
        X_set = X_set.reshape(shape[0], -1)
        X_set = np.hstack((X_set, np.ones((X_set.shape[0], 1))))
        A = np.linalg.lstsq(X_set, y_set, rcond=None)[0]
        
        self.intercept = A[-1]
        self.coef = A[:-1].T
        self.weight = self.coef * data
    
    
    def predict(self, data):
        '''
        利用近似的线性模型进行预测
        
        参数
        ----
        data：小批量输入数组，需要预测的数据
        
        
        Use approximate linear model to predict
        
        Parameter
        ---------
        data: barch input ndarray, input data
        '''
        if len(self.intercept.shape) == 0:
            result = data * self.coef
            retult = result.reshape(result.shape[0], -1)
            result = result.sum(axis=1) + self.intercept
            return result
        
        else:
            all_result = []
            for i in range(self.coef.shape[0]):
                result = data * self.coef[i]
                result = result.reshape(result.shape[0], -1)
                result = result.sum(axis=1) + self.intercept[i]
                all_result.append(result)
            return np.array(all_result).T
    
    
    def show(self, labels=None, index=None, top=None, estimate=None):
        '''
        作图并显示
        
        参数
        ----
        labels：一维列表，可选，特征名称，默认为None
        index：一维列表，可选，特征索引，默认为None，表示全选
        top：整型，可选，表示显示值最高的前top个特征，默认为None，表示全选
        estimate：整型，可选，表示图示示数保留的小数位数，默认为None
        
        
        Display the image
        
        Parameters
        ----------
        labels: 1-D list, callable, names of features, default=None
        index: 1-D list, callable, index of features, default=None, which means select all
        top: int, callable, display "top" features with the highest values, default=None, which means select all
        estimate: int, callable, indicating the number of decimal places reserved for the graphic display, default=None
        '''
        _image_process(self.weight, labels, index, top, estimate)
        plt.show()
    
    
    def savefig(self, filename, labels=None, index=None, top=None, estimate=None):
        '''
        作图并保存
        
        参数
        ----
        filename：字符串类型，文件名
        labels：一维列表，可选，特征名称，默认为None
        index：一维列表，可选，特征索引，默认为None，表示全选
        top：整型，可选，表示显示值最高的前top个特征，默认为None，表示全选
        estimate：整型，可选，表示图示示数保留的小数位数，默认为None
        
        
        Save the image
        
        Parameters
        ----------
        filename: str, file name
        labels: 1-D list, callable, names of features, default=None
        index: 1-D list, callable, index of features, default=None, which means select all
        top: int, callable, display "top" features with the highest values, default=None, which means select all
        estimate: int, callable, indicating the number of decimal places reserved for the graphic display, default=None
        '''
        _image_process(self.weight, labels, index, top, estimate)
        plt.savefig(filename)


'''
本模块储存着常用算符和积分方法
调用模块中的算符时应注意：
1、算符作用的多元函数的自变量需为列表(list、tuple、ndarray均可，下同)
2、算符作用的标量函数的输出值需为单个值
3、算符作用的矢量函数的输出值需为列表

This module stores the common operators and integration method
When calling operators in the module, pay attention to:
1. The argument of the multivariate function acted by the operators should be list (list, tuple, ndarray, the same below)
2. The output value of scalar function acted by the operators should be a single value
3. The output value of vector function acted by the operators should be list
'''

import numpy as np
from scipy import constants as C
from maysics.utils import grid_net
from matplotlib import pyplot as plt


def lim(f, x0, acc=0.01, method='both'):
    '''
    求极限
    lim x→x0+ f(x) ≈ f(x+dx)
    lim x→x0- f(x) ≈ f(x-dx)
    lim x→x0 f(x) ≈ (f(x+dx) + f(x-dx)) / 2
    
    参数
    ----
    f：函数类型
    x0：取极限的点
    acc：浮点数类型，可选，极限精度，即dx，默认为0.01
    method：字符串类型，求极限方法，可选'both'、'rigth'或'+'(右极限)、'left'或'-'(左极限)
    
    
    Calculate the limit values
    lim x→x0+ f(x) ≈ f(x+dx)
    lim x→x0- f(x) ≈ f(x-dx)
    lim x→x0 f(x) ≈ (f(x+dx) + f(x-dx)) / 2
    
    Parameters
    ----------
    f: function
    x0: point of limit
    acc: float, callable, the accuracy of calculation, equals to dx, default=0.01
    method: str, the method of calculation, 'both', 'right' or '+'(right limit), 'left' or '-'(left limit) are optional
    '''
    if method == 'both':
        func1 = f(x0 + acc)
        func2 = f(x0 - acc)
        return 0.5 * (func1 + func2)
    
    elif method == '+' or method == 'right':
        return f(x0 + acc)
    
    elif method == '-' or method == 'left':
        return f(x0 - acc)
    
    else:
        raise Exception("Parameter 'method' must be one of 'both', 'right', '+', 'left', '-'.")


def ha(f, m, U, acc=0.1):
    '''
    哈密顿算符：ha = - ħ**2 / 2m * ▽**2 + U
    
    参数
    ----
    f：函数
    m：数，粒子质量
    U：数或函数，势能
    acc：浮点数类型，可选，求导的精度，默认为0.1
    
    返回
    ----
    函数
    
    
    Hamilton: ha = - ħ**2 / 2m * ▽**2 + U
    
    Parameters
    ----------
    f: function
    m: num, the mass of the particle
    U: num or function, potential energy
    acc: float, callable, accuracy of derivation, default=0.1
    
    Return
    ------
    function
    '''
    def obj(x):
        x = np.array(x, dtype=float)
        result = 0
        for i in range(len(x)):
            func1 = 2 * f(x)
            x[i] += acc
            func2 = f(x)
            x[i] -= 2 * acc
            func3 = f(x)
            x[i] += acc
            de = (func2 + func3 - 2 * func1) / acc**2
            result += de
        result *= (C.h / (2 * np.pi))**2 / (2 * m)
        
        if type(U).__name__ == 'function':
            result += U(x) * f(x)
        else:
            result += U * f(x)
        return result
    return obj


def grad(f, x, acc=0.1):
    '''
    求标量函数梯度：▽f(x)
    
    参数
    ----
    f：函数，要求函数f返回一个数值
    x：数或数组，函数的输入值，不支持批量输入
    acc：浮点数类型，可选，求导的精度，默认为0.1
    
    返回
    ----
    梯度函数值 ▽f(x)
    
    
    gradient of scalar function: ▽f(x)
    
    Parameters
    ----------
    f: function, the function should return a number
    x: num or array, the input of the function, batch input is not supported
    acc: float, callable, accuracy of derivation, default=0.1
    
    Return
    ------
    value of gradient function ▽f(x)
    '''
    x = np.array(x, dtype=float)
    d_list = []
    acc2 = 0.5 * acc
    if len(x.shape) == 0:
        x += acc2
        func1 = f(x)
        x -= acc
        func2 = f(x)
        x += acc2
        d_list = (func1 - func2) / (acc)
    
    elif len(x.shape) == 1:
        for i in range(x.shape[0]):
            x[i] += acc2
            func1 = f(x)
            x[i] -= acc
            func2 = f(x)
            de = (func1 - func2) / (acc)
            x[i] += acc2
            d_list.append(de)
        d_list = np.array(d_list)
    
    elif len(x.shape) == 2:
        for i in range(x.shape[1]):
            x[0][i] += acc2
            func1 = f(x)
            x[0][i] -= acc
            func2 = f(x)
            de = (np.array([func1]).T - np.array([func2]).T) / (acc)
            x[0][i] += acc2
            d_list.append(de)
        d_list = [d_list]
        d_list = np.array(d_list)
    
    return d_list


def nebla_dot(f, x, acc=0.1):
    '''
    ▽点乘矢量函数：▽·f(x)
    
    参数
    ----
    f：函数，要求函数f返回一个列表
    x：数或数组，函数的输入值，不支持批量输入
    acc：浮点数类型，可选，求导的精度，默认为0.1
    
    返回
    ----
    函数值 ▽·f(x)
    
    
    dot product between ▽ and vector function: ▽·f(x)
    
    Parameter
    ---------
    f: function, the function should return a list
    x: num or array, the input of the function, batch input is not supported
    acc: float, callable, accuracy of derivation, default=0.1
    
    Return
    ------
    value of function ▽·f(x)
    '''
    x = np.array(x, dtype=float)
    result = []
    if len(x.shape) == 1:
        for i in range(x.shape[0]):
            x[i] += acc * 0.5
            func1 = f(x)[i]
            x[i] -= acc
            func2 = f(x)[i]
            x[i] += acc * 0.5
            de = (func1 - func2) / acc
            result.append(de)
    else:
        for i in range(x.shape[1]):
            x[0][i] += acc * 0.5
            func1 = f(x)[0][i]
            x[0][i] -= acc
            func2 = f(x)[0][i]
            x[0][i] += acc * 0.5
            de = (func1 - func2) / acc
            result.append(de)
        result = [result]
    result = np.array(result)
    return result


def _diff2forc(f, x, a, b, acc):
    if len(x.shape) == 1:
        x[b] += acc * 0.5
        func1 = f(x)[a]
        x[b] -= acc
        func2 = f(x)[a]
        x[b] += acc * 0.5
        de = (func1 - func2) / acc
    else:
        x[0][b] += acc * 0.5
        func1 = f(x)[0][a]
        x[0][b] -= acc
        func2 = f(x)[0][a]
        x[0][b] += acc * 0.5
        de = (func1 - func2) / acc
    return de
    
    
def nebla_cross(f, x, acc=0.1):
    '''
    ▽叉乘矢量函数：▽×f(x)
    
    参数
    ----
    f：函数，要求函数是三维矢量函数
    x：数或数组，函数的输入值，不支持批量输入
    acc：浮点数类型，可选，求导的精度，默认为0.1
    
    返回
    ----
    函数值 ▽×f(x)
    
    
    ▽ cross vector function: ▽×f(x)
    
    Parameter
    ---------
    f: function, the function should be a three-dimension vector function
    x: num or array, the input of the function, batch input is not supported
    acc: float, callable, accuracy of derivation, default=0.1
    
    Return
    ------
    value of function ▽×f(x)
    '''
    x = np.array(x, dtype=float)
    result = np.array([
        _diff2forc(f, x, 2, 1, acc) -\
        _diff2forc(f, x, 1, 2, acc),
        _diff2forc(f, x, 0, 2, acc) -\
        _diff2forc(f, x, 2, 0, acc),
        _diff2forc(f, x, 1, 0, acc) -\
        _diff2forc(f, x, 0, 1, acc)])
    return result


def laplace(f, x, acc=0.1):
    '''
    △算子 = ▽**2
    被作用函数的自变量x是列表
    
    参数
    ----
    f：函数
    x：一维数组或二维数组
    acc：浮点数类型，可选，求导的精度，默认为0.1
    
    返回
    ----
    函数值△f(x)
    
    
    Laplace operator: △ = ▽**2
    the argument x of the affected function should be a list
    
    Parameter
    ---------
    f: function
    x: 1-D or 2-D array
    acc: float, callable, accuracy of derivation, default=0.1
    
    Return
    ------
    value of function △f(x)
    '''
    x = np.array(x, dtype=float)
    result = 0
    if len(x.shape) == 1:
        for i in range(x.shape[0]):
            func1 = 2 * f(x)
            x[i] += acc
            func2 = f(x)
            x[i] -= 2 * acc
            func3 = f(x)
            x[i] += acc
            de = (func2 + func3 - func1) / acc**2
            result += de
    else:
        for i in range(x.shape[1]):
            func1 = 2 * f(x)
            x[:, i] += acc
            func2 = f(x)
            x[:, i] -= 2 * acc
            func3 = f(x)
            x[:, i] += acc
            de = (func2 + func3 - func1) / acc**2
            result += de
    
    return result


def _mc_fit(func, condition, random_state, area, dim, args, param, loop, height):
    '''
    蒙特卡洛法积分
    
    
    MonteCarlo method
    '''
    np.random.seed(random_state)
    area = np.array(area)
    area_length = area[:, 1] - area[:, 0]
    V = area_length.prod() * height
    
    num = len(area_length)
    area_points = np.random.rand(loop, num) * area_length + area[:, 0]
    func_points = np.random.rand(loop) * height
    
    effect_points = 0
    if dim == 1:
        if not condition:
            for i in range(loop):
                if func_points[i] <= func(area_points[i], **args):
                    effect_points += 1
        
        else:
            for i in range(loop):
                if func_points[i] <= func(area_points[i], **args) and condition(area_points[i], **param):
                    effect_points += 1
    
    elif dim == 2:
        func_points_2 = func(area_points, **args)
        if not condition:
            for i in range(loop):
                if func_points[i] <= func_points_2[i]:
                    effect_points += 1
        
        else:
            for i in range(loop):
                if func_points[i] <= func_points_2[i] and condition(area_points[i], **param):
                    effect_points += 1
    
    else:
        raise Exception("Parameter 'dim' must be one of 1 and 2.")
    
    effect_points_rate = effect_points / loop
    
    return V * effect_points_rate


def _rect_fit(func, condition, area, dim, args, param, acc):
    '''
    矩形法积分
    
    
    Rectangle method
    '''
    dv = acc.prod()
    
    points_net = []
    for i in range(len(area)):
        points_net.append(np.arange(area[i][0], area[i][1], acc[i]))
    points_net = grid_net(*points_net)
    
    if dim == 1:
        func_points=[]
        if not condition:
            for i in points_net:
                func_points.append(func(i, **args))
        
        else:
            for i in points_net:
                if condition(i, **param):
                    func_points.append(func(i, **args))
        
        func_points = np.array(func_points) * dv
    
    elif dim == 2:
        points_net = points_net.tolist()
        if condition:
            for i in points_net[:]:
                if not condition(i, **param):
                    points_net.remove(i)
        points_net = np.array(points_net)
        func_points = func(points_net, **args) * dv
    
    else:
        raise Exception("Parameter 'dim' must be one of 1 and 2.")
    
    return func_points.sum()


def inte(func, area, method='rect', dim=1, args={}, condition=None, param={}, acc=0.1, loop=10000, height=1, random_state=None):
    '''
    定积分
    
    矩形法rect：
    S ≈ ∑ f(xi)dv
    dv = ∏ dx ≈ ∏ Δx = ∏ acc
    
    蒙特卡洛法mc：
    在区域V = area * height中产生loop个随机点
    计算符合条件的随机点个数effect_points
    S ≈ V * effect_points / loop
    
    参数
    ----
    func：函数类型，被积函数
    area：二维列表，积分区域，由自变量上下限列表组成
        如：积分区域为[a, b]时，area=[[a, b]]
            积分区域为二维区域x1∈[a1, b1]，x2属于[a2, b2]时，area=[[a1, b1], [a2, b2]]
    method：字符串类型，可选'rect'、'mc'，默认为'rect'
    dim：整型，可选1或2，1表示被积函数的输入为1维数组，适用于普通输入函数，2表示被积函数的输入为2维数组，适用于小批量输入函数，默认为1
    args：字典类型，可选，当被积函数有其他非默认参数时，需输入以参数名为键，参数值为值的字典，默认为空字典
    condition：函数类型，可选，条件函数，符合条件的输出Ture，否则输出False，条件函数的第一个参数的输入须为1维数组
    param：字典类型，可选，当条件函数有其他非默认参数时，需输入以参数名为键，参数值为值的字典，默认为空字典
    acc：浮点数或列表类型，可选，积分精度，只有method='rect'时才有效，默认为0.1
    loop：整型，可选，产生的随机数数目，只有method='mc'时才有效，默认为10000
    height：浮点数类型，可选，高度，只有method='mc'时才有效，默认为1
    random_state：整型，随机种子，只有method='mc'时才有效
    
    
    Definite integral
    
    ractangle method ('rect')：
    S ≈ ∑ f(xi)dv
    dv = ∏ dx ≈ ∏ Δx = ∏ acc
    
    MonteCarlo method ('mc')：
    generate 'loop' random points in area 'V = area * height'
    calculate the number of qualified points 'effect_points'
    S ≈ V * effect_points / loop
    
    Parameters
    ----------
    func: function, integrand
    area: 2-D list, integral region, composed of a list of upper and lower limits of independent variables
        e.g. when the integral region is [a, b], area = [[a, b]]
             when the 2-D integral region is x1∈[a1, b1] and x2∈[a2, b2], area=[[a1, b1], [a2, b2]]
    method: str, callable, 'rect' and 'mc' are optional, default='rect'
    dim: int, 1 and 2 are optional, 1 means the input of integrand is 1-D list, like normal functions, 2 means the input of integrand is 2-D list, like functions need mini-batch input, default=1
    args: dict, callable, when integrand function has other non-default parameters, "args" needs to be input a dictionary with parm_name as key and param_value as value, an empty dict to default
    condition: function, callable, condition function with the input of the first parameter as 1-D list, if input if qualified, ouput True, otherwise output False
    param: dict, callable, when condtition function has other non-default parameters, "param" needs to be input a dictionary with parm_name as key and param_value as value, default={}
    acc: float or list, callable, integration accuracy, it's effective only when method='rect', default=0.1
    loop: int, callable, the number of generated random numbers, it's effective only when method='mc', default=10000
    height: float, callable, height, it's effective only when method='mc', default=1
    random_state: int, random seed, it's effective only when method='mc'
    '''
    area = np.array(area, dtype=float)
    if method == 'rect':
        if type(acc).__name__ == 'float' or type(acc).__name__ == 'int':
            acc = np.ones(len(area)) * acc
        else:
            acc = np.array(acc)
        return _rect_fit(func, condition, area, dim, args, param, acc)
    
    elif method == 'mc':
        return _mc_fit(func, condition, random_state, area, dim, args, param, loop, height)
    
    else:
        raise Exception("Parameter 'method' must be one of 'rect' and 'mc'.")